/*--------------------------------------------------------------------------------------*/
/* Text -> FCTS Converter                                                               */
/* Programmer: Austin Chu (qarv@yahoo.com)                                              */
/* Programmed in C for the lcc-win32 compiler with -ANSI option                         */
/* Version: 0.0.2 (alpha-2, bug fix for creating filenames)                             */
/*                                                                                      */
/* txt2fcts.c - for now, the whole program                                              */
/*--------------------------------------------------------------------------------------*/

/* Note:    This program takes a file with flash card information:
                    card1 front<endl>
                    card1 back<endl>
                    card2 front<endl>
                    ...
                        <endl> == any succession of 0x0A's and/or 0x0D's
                        terminated with or without an <endl>
            and converts it into an *.89y file which contains an FCTS input file for
            transfer by link cable to the TI89. */

/* Second note: The procedures for creating a file acceptable to the TI-Graph Link have
                been adapted from the excellent ttbin2oth program in TI-Chess Team's TIGCC
                Tool Suite version 0.98.  Their website is http://tict.ticalc.org/. */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#define THIS_VERSION 0x0002                 /* The version number of this compiliation */

void PrintSyntax(void);
unsigned char* Text2Data(char* inputfile, int* size);
unsigned char* Create89y(char* tifolder, char* varname,
                         unsigned char* inputdata, int inputsize, int* outputsize);

/* Shall we crunch some flash cards? */
int main(int argc, char *argv[])
{
    char* outpath = NULL;                   /* The file to output to */
    char tifolder[9] = "";                  /* The destination folder on the TI89 */
    char varname[9] = "";                   /* The destination variable name on the TI89 */
    FILE* ofstream;                         /* The output file stream */
    unsigned char* inputdata;               /* The input data for the fcts file */
    int inputsize;                          /* The size of the input data */
    unsigned char* outputdata;              /* The output data for the *.89y file */
    int outputsize;                         /* The size of the output data */
    int i;                                  /* A loop counter */
    int j;                                  /* A second counter */

    printf("Text -> FCTS Converter v. 0.0.2 by Austin Chu\n");

    if(argc == 1)
    {
        PrintSyntax();
        return 1;
    }

    printf("Preparing for conversion...\n");

    /* Gather the switches (argv[0] is txt2fcts and argv[1] is the input filename) */
    for(i = 2; i < argc; i++)
        /* Is the parameter a proper switch? */
        if(argv[i][0] == '-' && argv[i][2] == ':')
            if(argv[i][1] == 'o')
            {
                /* Allocate just enough memory for the output folder size */
                if(!(outpath = (char*)realloc(outpath, strlen(&argv[i][3]) + 14)))
                {
                    printf(
                   "ERROR: txt2fcts could not allocate the required memory (%u bytes).\n",
                           strlen(&argv[i][3]) + 14);
                    return 1;
                }

                strcpy(outpath, &argv[i][3]);
            }
            else if(argv[i][1] == 'f')
                strcpy(tifolder, &argv[i][3]);
            else if(argv[i][1] == 'v')
                strcpy(varname, &argv[i][3]);
            else
            {
                if(outpath)
                    free(outpath);
                printf("ERROR: txt2fcts encountered an incorrect parameter (%s).\n",
                       argv[i]);
                PrintSyntax();
                return 1;
            }
        else
        {
            if(outpath)
                free(outpath);
            printf("ERROR: txt2fcts encountered an incorrect parameter (%s).\n", argv[i]);
            PrintSyntax();
            return 1;
        }

    /* Set up the default destination folder */
    if(!tifolder[0])
        strcpy(tifolder, "main");

    /* Set up the default variable name */
    if(!varname[0])
    {
        j = 0;

        /* Find the end of folder listings */
        i = strlen(argv[1]);
        while(argv[1][--i] != '\\')
            if(!i)
            {                               /* We've hit the beginning of the string */
                i--;                        /* Fixed v0.0.2; bug found by Gatlin Hardin */
                break;
            }

        while(j < 8 && argv[1][++i])        /* Loop until full or end of input filename */
                /* A number cannot start a variable name */
            if(argv[1][i] >= '0' && argv[1][i] <= '9' && j != 0)
                varname[j++] = argv[1][i];
                /* Check whether the character is a letter */
            else if((argv[1][i] >= 'A' && argv[1][i] <= 'Z') ||
                    (argv[1][i] >= 'a' && argv[1][i] <= 'z'))
                varname[j++] = argv[1][i];
                            /* Stop if we reach the file extension */
            else if(argv[1][i] == '.')
                break;
    }

    if(!varname[0])                         /* Nothing qualified; set up the default */
        strcpy(varname, "cards");

    printf("Reading the input file...\n");
    if(!(inputdata = Text2Data(argv[1], &inputsize)))
    {
        if(outpath)
            free(outpath);
        return 1;
    }

    printf("Creating the output data...\n");
    if(!(outputdata = Create89y(tifolder, varname, inputdata, inputsize, &outputsize)))
    {
        if(outpath)
            free(outpath);
        free(inputdata);
        return 1;
    }

    free(inputdata);

    printf("Writing to the output file...\n");

    /* Get the output file name and then open it */
    if(!outpath)
    {                                       /* There wasn't an output folder specified */
        /* Find the end of folder listings */
        i = strlen(argv[1]);
        while(argv[1][--i] != '\\')
            if(!i)
                break;

        if(!i)
        {
            if(!(outpath = (char*)malloc(13)))
            {
                free(outputdata);
                printf(
                  "ERROR: txt2fcts could not allocate the required memory (13 bytes).\n");
                return 1;
            }
            memset(outpath, 0, 13);
        }
        else
        {
            if(!(outpath = (char*)malloc(i + 14)))
            {
                free(outputdata);
                printf(
                   "ERROR: txt2fcts could not allocate the required memory (%u bytes).\n",
                       i + 14);
                return 1;
            }
            for(j = 0; j <= i; j++)
                outpath[j] = argv[1][j];
            memset(outpath + i + 1, 0, 13);
        }
    }
    else
        strcat(outpath, "\\");
    strcat(strcat(outpath, varname), ".89y");
    if(ofstream = fopen(outpath, "rb"))
    {                                       /* The output file already exists */
        printf("Are you sure you want to overwrite %s? (y/n) ", outpath);
        j = getchar();
        if(j != 'y' && j != 'Y')
        {
            free(outpath);
            free(outputdata);
            fclose(ofstream);
            printf("txt2fcts operation halted by the user.\n");
            return 1;
       }
    }
    if(!(ofstream = fopen(outpath, "wb")))
    {
        free(outputdata);
        printf("ERROR: txt2fcts could not open the output file %s.\n", outpath);
        free(outpath);
        return 1;
    }

    if(fwrite(outputdata, outputsize, 1, ofstream) != 1)
    {
        fclose(ofstream);
        remove(outpath);
        free(outputdata);
        printf("ERROR: txt2fcts could not write %u bytes to output file %s.\n",
               outputsize, outpath);
        free(outpath);
        return 1;
    }

    printf("txt2fcts has successfully written %u bytes to output file %s.\n",
           outputsize, outpath);

    fclose(ofstream);
    free(outputdata);
    free(outpath);

    return 0;
}

/* Displays the proper syntax for a call to txt2fcts in case of syntaxical error
   Arguments: none
   Returns: nothing */
void PrintSyntax(void)
{
    printf("Proper syntax:\n");
    printf("txt2fcts <input_file> [switches]\n");
    printf("input_file: the path to the input text file\n");
    printf("switches:   -o:<folder>  -- send the *.89y file to a certain folder\n");
    printf("                                default --> folder of the input file\n");
    printf("            -f:<folder>  -- specify the destination folder on the TI89\n");
    printf("                                default --> main\n");
    printf("            -v:<varname> -- specify a variable name for the new fcts file\n");
    printf("                                default --> first eight alpha-numeric\n");
    printf("                                            characters in input_file\n");

    return;
}

/* Reads from the input file and creates the input data for the fcts file
   Arguments: the filename of the input file
              a pointer to an integer in which the size of the data is stored
   Returns: a pointer to the dynamically allocated data */
unsigned char* Text2Data(char* inputfile, int* size)
{
    FILE* ifstream;                         /* The input file stream */
    unsigned char* data;                    /* The data buffer */
    int i = 0;                              /* The counter for the data buffer */
    int linecount = 0;                      /* The number of lines in the file */

    /* Open the input file */
    if(!(ifstream = fopen(inputfile, "rb")))
    {
        printf("ERROR: txt2fcts could not open the input file %s.\n", inputfile);
        return NULL;
    }

    /* Allocate maximum possible buffer considering the input file size */
    fseek(ifstream, 0, SEEK_END);
    if(!(data = (unsigned char*)malloc(ftell(ifstream) + 6)))
    {
        fclose(ifstream);
        printf("ERROR: txt2fcts could not allocate the required memory (%u bytes).\n",
               ftell(ifstream) + 6);
        return NULL;
    }
    rewind(ifstream);

    /* Write the format version */
    data[0] = (unsigned char)((THIS_VERSION >> 8) & 0xff);
    data[1] = (unsigned char)(THIS_VERSION & 0xff);

    data[4] = 0x00;                         /* The front buffer null byte */

    i = 5;

    /* Read the input file */
    while(fread(data + i, sizeof(data[0]), 1, ifstream) == 1)
        if(data[i] == 0x0A || data[i] == 0x0D)
        {                                   /* Skip multiple endl's */
            linecount++;

            do
                if(fread(data + i, sizeof(data[0]), 1, ifstream) != 1)
                    data[i] = 0x00;         /* We've hit the end of the file */
            while(data[i] == 0x0A || data[i] == 0x0D);

            if(data[i++] == 0x00)
                break;                      /* We've hit the end of the file */

            /* Insert the space for the null byte */
            data[i] = data[i - 1];
            data[i++ - 1] = 0x00;
        }
        else
            i++;

    fclose(ifstream);

    /* Check whether the file terminated with endl */
    if(data[i - 1] != 0x00)
    {
        linecount++;
        data[i++] = 0x00;
    }

    *size = i;                              /* Record the size */

    if(linecount % 2 == 1)
    {                                       /* Odd number of data lines */
        free(data);
        printf("ERROR: The input file contains an odd number of data lines.\n");
        return NULL;
    }

    i = linecount / 2;
    data[2] = (unsigned char)((i >> 8) & 0xff);
    data[3] = (unsigned char)(i & 0xff);

    /* Re-allocate to fit our new needs */
    if(!(data = realloc(data, *size)))
    {
        printf("ERROR: txt2fcts could not allocate the required memory (%u bytes).\n",
               size);
        return NULL;
    }

    return data;
}

/* Creates the data to be put into the *.89y file
   (Note: The following function is based largely on DataBuffer2OTHBuffer() in bin2oth.c
          in the TIGCC Tools Suite v0.98 by the TI-Chess Team (http://tict.ticalc.org).)
   Arguments: the destination folder on the calculator
              the destination variable name on the calculator
              the input data to be sent to the calculator
              the size of the input data
              a pointer to a place to store the output size
   Returns: a pointer to the dynamically allocated output data */
unsigned char* Create89y(char* tifolder, char* varname,
                         unsigned char* inputdata, int inputsize, int* outputsize)
{
    /* The definition of the header for the *.89y file */
    typedef struct
    {
        char calculator[8];                 /* "**TI89**" */
        unsigned char fill1[2];             /* 01 00 */
        char folder[8];                     /* The destination folder name */
        char desc[40];                      /* Not used */
        unsigned char fill2[6];             /* 01 00 52 00 00 00 */
        char name[8];                       /* The destination variable name */
        unsigned char type[4];              /* 0C 00 00 00 */
        unsigned char size[4];              /* Complete file size (including checksum) */
        unsigned char fill3[6];             /* A5 5A 00 00 00 00 */
        unsigned char datasize[2];          /* The data size */
    } Header;

    Header* theHeader;                      /* The header structure in output */
    unsigned char* ptr;                     /* A traveling pointer */
    int size;                               /* A temporary holder for sizes */
    unsigned char* output;                  /* The output buffer */
    unsigned int checksum;                  /* The checksum */

    if(inputsize > 65519)
    {                                       /* The input data is too big */
        printf("ERROR: The input data is bigger than the possible 65519 bytes.\n");
        return NULL;
    }

    size = sizeof(Header) + inputsize + 9;  /* 7 OTH marker bytes and 2 checksum bytes */

    if(!(output = (unsigned char*)malloc(size)))
    {
        printf("ERROR: txt2fcts could not allocate the required memory (%u bytes).\n",
               size);
        return NULL;
    }

    /* Set up the header */
    *outputsize = size;
    theHeader = (Header*)output;
    memset(theHeader,0,sizeof(Header));

    /* Fill the static fields that aren't null*/
    theHeader->fill1[0] = 0x01;
    theHeader->fill2[0] = 0x01; theHeader->fill2[2] = 0x52;
    theHeader->fill3[0] = 0xA5; theHeader->fill3[1] = 0x5A;
    theHeader->type[0] = 0x0c;  theHeader->type[2] = 0x03;

    /* Fill the calculator type indicator */
    strncpy(theHeader->calculator, "**TI89**", 8);

    /* Fill in the destination folder and variable name*/
    strncpy(theHeader->folder, tifolder, 8);
    strncpy(theHeader->name, varname, 8);

    /* Fill in the size field with size of entire file */
    theHeader->size[0] = (unsigned char)(size & 0xff);
    theHeader->size[1] = (unsigned char)((size >> 8)  & 0xff);
    theHeader->size[2] = (unsigned char)((size >> 16) & 0xff);
    theHeader->size[3] = (unsigned char)((size >> 24) & 0xff);

    /* Fill data size with size of just the data (w/ OTH tag bytes) */
    size = inputsize + 7;
    theHeader->datasize[0] = (unsigned char)((size >> 8) & 0xff);
    theHeader->datasize[1] = (unsigned char)(size & 0xff);

    /* Copy the input data */
    memcpy(output + sizeof(Header), inputdata, inputsize);

    /* Add the fcts extension and the OTH tag */
    ptr    = output + sizeof(Header) + inputsize;
    *ptr++ = 0;
    strcpy(ptr, "fcts");
    ptr += 5;
    *ptr++ = 0xF8;                          /* The OTH tag */

    /* Accumulate and record the checksum */
    size = inputsize + 9;
    ptr = theHeader->datasize;
    checksum = 0;
    while(size--)
        checksum += *ptr++;
    output[*outputsize - 2] = (unsigned char)(checksum & 0xff);
    output[*outputsize - 1] = (unsigned char)((checksum >> 8) & 0xff);

    return output;
}